use std::collections::HashMap;

use lazy_static::lazy_static;
use regex::Regex;


struct Directory {
    name: String,
    children: Vec<Node>,
}


struct File {
    name: String,
    size: usize,
}


enum Node {
    Branch(Directory),
    Leaf(File),
}



lazy_static! {
    static ref CMD_CD: Regex = Regex::new(r"^\$ cd (\S+)$").unwrap();
    static ref CMD_LS: Regex = Regex::new(r"^\$ ls$").unwrap();
    static ref DIR_LISTING: Regex = Regex::new(r"^dir (\S+)$").unwrap();
    static ref FILE_LISTING: Regex = Regex::new(r"^(\d+) (\S+)$").unwrap();
}


struct DirListing {
    name: String,
    children: Vec<usize>,
    parent: Option<usize>,
}

enum NodeListing {
    Branch(DirListing),
    Leaf(File),
}


fn build_map(data: &str) -> Result<HashMap<usize, NodeListing>, String> {
    let root = DirListing {
        name: "".to_string(),
        children: vec![],
        parent: None,
    };
    let mut map = HashMap::new();
    map.insert(0, NodeListing::Branch(root));

    let mut stack: Vec<usize> = vec![];
    stack.push(0);

    'lines: for line in data.trim().lines() {
        if CMD_LS.is_match(line) {continue;}

        if let Some(c) = CMD_CD.captures(line) {
            let name = c.get(1).unwrap().as_str();
            if name == "/" {
                stack.truncate(1);
                continue;
            }

            if name == ".." {
                stack.pop();
                continue;
            }

            let current_dir = match map.get(&stack.last().unwrap()).unwrap() {
                NodeListing::Branch(dir) => dir,
                _ => unreachable!(),
            };
            for id in &current_dir.children {
                let child_name = match map.get(&id).unwrap() {
                    NodeListing::Branch(dir) => &dir.name,
                    NodeListing::Leaf(file) => &file.name,
                };
                if child_name == name {
                    stack.push(*id);
                    continue 'lines;
                }
            }

            return Err(format!("Attempted to cd into directory before ls-ing it: {}", line.to_string()));
        }

        if let Some(c) = DIR_LISTING.captures(line) {
            let parent_id = stack.last().unwrap();
            let child_id = map.len();
            let parent = match map.get_mut(parent_id).unwrap() {
                NodeListing::Branch(dir) => dir,
                _ => unreachable!(),
            };
            parent.children.push(child_id);

            let dir = DirListing {
                name: c.get(1).unwrap().as_str().to_string(),
                children: vec![],
                parent: Some(*parent_id),
            };
            map.insert(child_id, NodeListing::Branch(dir));
            continue;
        }

        if let Some(c) = FILE_LISTING.captures(line) {
            let name = c.get(2).unwrap().as_str().to_string();
            let size = c.get(1).unwrap().as_str().parse::<usize>()
                .map_err(|_e| format!("Invalid file size: {}", line.to_string()))?;

            let file_id = map.len();
            let file = File {name, size};
            map.insert(file_id, NodeListing::Leaf(file));

            let parent = match map.get_mut(stack.last().unwrap()).unwrap() {
                NodeListing::Branch(dir) => dir,
                _ => unreachable!(),
            };
            parent.children.push(file_id);
            continue;
        }

        return Err(format!("Could not interpret line: {line}"));
    }

    Ok(map)
}


fn build_tree(map: &mut HashMap<usize, NodeListing>, root_listing: &DirListing) -> Directory {
    let mut root_children = Vec::new();
    for child_id in &root_listing.children {
        match map.remove(&child_id).unwrap() {
            NodeListing::Leaf(file) => {
                root_children.push(Node::Leaf(file));
            },
            NodeListing::Branch(dir_listing) => {
                let dir = build_tree(map, &dir_listing);
                root_children.push(Node::Branch(dir))
            },
        }
    }

    Directory {
        name: root_listing.name.clone(),
        children: root_children,
    }
}


fn get_sizes(root: &Directory, mut sizes: Vec<usize>) -> (usize, Vec<usize>) {
    let mut dir_size = 0;

    for child in &root.children {
        match child {
            Node::Leaf(file) => dir_size += file.size,
            Node::Branch(dir) => {
                let (child_size, rs) = get_sizes(dir, sizes);
                sizes = rs;
                dir_size += child_size;
            },
        }
    }

    sizes.push(dir_size);
    (dir_size, sizes)
}


fn main() -> Result<(), String> {
    let mut map = build_map(DATA)?;
    let root = match map.remove(&0).unwrap() {
        NodeListing::Branch(dir) => dir,
        _ => unreachable!(),
    };

    let tree = build_tree(&mut map, &root);
    let (total_size, dir_sizes) = get_sizes(&tree, vec![]);

    let part1_sizes: usize = dir_sizes.iter()
        .filter(|&&s| s <= 100000)
        .sum();
    println!("Part 1: {part1_sizes}");

    let available = 70_000_000 - total_size;
    if available >= 30_000_000 {
        return Err("Weird, there's already enough space available".to_string());
    }
    let target_size = 30_000_000 - available;

    let delete_size = dir_sizes.iter()
        .filter(|&&s| s >= target_size)
        .min()
        .unwrap();
    println!("Part 2: {delete_size}");

    Ok(())
}


const DATA: &'static str = include_str!("../data/day7.txt");
// const DATA: &'static str = "
// $ cd /
// $ ls
// dir a
// 14848514 b.txt
// 8504156 c.dat
// dir d
// $ cd a
// $ ls
// dir e
// 29116 f
// 2557 g
// 62596 h.lst
// $ cd e
// $ ls
// 584 i
// $ cd ..
// $ cd ..
// $ cd d
// $ ls
// 4060174 j
// 8033020 d.log
// 5626152 d.ext
// 7214296 k
// ";
